# -*- coding: utf-8 -*-
#    Direct closeness of fit to Zipf's law based on Kolgomoroff-Smirnof
#    distance, using first derivate-based exploration of parameter values
#    Copyright (C) 2009-2011 - Philippe Aigrain
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from pylab import *
import csv

# Input data format : a CSV file containing 1-uples in decreasing order
# Number in line i : access level (integer) for i^th ranked
# Any following numbers in a line will be ignored
# It is assumed that all ranks are present (exhaustive data).
# If only partial data is available, rather use the MLE estimate
# method described in Michel L. Goldstein, Steven A. Morris, and Gary G. Yen.
# ‘Problems with fitting to the power-law distribution’.
# The European Physical Journal B, 41(2), September 2004.
# preprint from http://arxiv.org/abs/cond-mat/0402322.

def harmonic(s,n):
  k=0.0;
  for i in arange (1.0,n+1.0,1.0):
    k += pow(i,-s);
  return k;  

def main(argv):
  if len(argv) < 2:
    print ("Usage: %s <CSV file containing decresing order access level (integer)>") % argv[0]
    print "Kolmogorov-Smirnof distance-based dynamic approximation of fit on cumulative distributions"
    sys.exit(0)
  # Let's first compute the number of lines in the data file
  cr = csv.reader(open(argv[1],"rb"))
  N = 0
  for row in cr:
    N += 1
  # Parameter defining the lowest difference in parameter of Zipf's law we consider meaningful
  # Will be used for both derivatives and stopping approximation
  threshold = 0.00001
  data = reshape(zeros(N*2,float),(N,2))
  # Reading data
  cr = csv.reader(open(argv[1],"rb"))
  sumobs = 0.0
  i = 0
  for row in cr:
    data[i,0] = 1.0*(i+1)
    data[i,1] = 1.0*(int) (row[0])
    sumobs += data[i,1]
    i += 1
  # Computing once and for all the observed cumulated distribution
  observed = zeros(N,float)
  cumobs = zeros (N,float)
  cumobs[0] = data[0,1] / sumobs 
  for i in range(N-1):
    observed[i] = data[i+1,1]/sumobs
    cumobs[i+1] = cumobs[i]+data[i+1,1]/sumobs
  print "total access: %d" % sumobs
  # Looking for a decent starting point for parameter within [0.1-1.5]
  modelled = zeros(N,float)
  cummod = zeros (N,float)
  va = zeros(15, float)
  for j in range(15):
    a = 0.1 * (j+1)
    # Computing the modelled CDF and the Kolmogorov-Smirnof distance
    ha0N = harmonic(a,N)
    modelled[0] = pow(1.0,-a)/ha0N
    cummod[0] = pow(1.0,-a)/ha0N
    maxdist = abs(cumobs[0]-cummod[0])
    for i in arange(1.0,N-1,1.0):
      modelled[i] = pow(i+1.0,-a)/ha0N     
      cummod[i] = cummod[i-1]+(pow(i+1.0,-a)/ha0N)
      if maxdist < abs(cumobs[i]-cummod[i]):
        maxdist = abs(cumobs[i]-cummod[i])
    va[j] = maxdist
  # Initialization (look for minimum value)
  mindist = 1.0
  for j in range(15):
    if va[j] < mindist:
      mindist = va[j]
      newa1 = 0.1 * (j+1)
      inewa1 = j
  if inewa1 == 0:
    a1 = newa1
    ks1 = va[inewa1]
    a2 =  a1 + 0.1
    ks2 = va[inewa1+1]
  elif inewa1 >= 13:
    a1 = newa1 - 0.1
    ks1 = va[inewa1-1]
    a2 = 1.5
    ks2 = va[14]
  else:
    a1 = newa1 - 0.1
    ks1 = va[inewa1-1]
    a2 = newa1 + 0.1
    ks2 = va[inewa1+1]
  progress = 1.0
  step = 0
  print "step: %d, newa1: %f, ks1: %f" %(0,newa1,ks1)
  # Iteration
  while progress > (10.0*threshold):
    step += 1
    # Computing first derivatives at a1 and a2
    da1 = a1 + threshold
    da2 = a2 + threshold
    print "da1: %f, da2: %f" %(da1,da2)
    # Computing modelled distribution and KS distance at da1 and da2
    # computation for da1
    ha0N = harmonic(da1,N)
    modelled[0] = pow(1.0,-da1)/ha0N
    cummod[0] = pow(1.0,-da1)/ha0N
    maxdist = abs(cumobs[0]-cummod[0])
    for i in arange(1.0,N-1,1.0):
      modelled[i] = pow(i+1.0,-da1)/ha0N     
      cummod[i] = cummod[i-1]+(pow(i+1.0,-da1)/ha0N)
      if maxdist < abs(cumobs[i]-cummod[i]):
        maxdist = abs(cumobs[i]-cummod[i])
    ksd1 = maxdist
    # computation for da2
    ha0N = harmonic(da2,N)
    modelled[0] = pow(1.0,-da2)/ha0N
    cummod[0] = pow(1.0,-da2)/ha0N
    maxdist = abs(cumobs[0]-cummod[0])
    for i in arange(1.0,N-1,1.0):
      modelled[i] = pow(i+1.0,-da2)/ha0N     
      cummod[i] = cummod[i-1]+(pow(i+1.0,-da2)/ha0N)
      if maxdist < abs(cumobs[i]-cummod[i]):
        maxdist = abs(cumobs[i]-cummod[i])
    ksd2 = maxdist
    y1 = (ksd1-ks1)/threshold
    y2 = (ksd2-ks2)/threshold
    # If both derivatives have same sign change interval, if not dig inside interval
    if ((y1*y2)>0.0):
      #look for direction
      if y1 < 0.0:
	# we look higher
	newa1 = max(a1,a2)
      else:
	# we look below
	newa1 = min(a1,a2)
    else:
      # we dig into the interval
      newa1 = ((ks2-ks1)+(y1*a1)-(y2*a2))/(y1-y2)
      progress = min(abs(newa1-a1),abs(newa1-a2))
    print "ksd1: %f, ksd2: %f, y1: %f, y2: %f, newa1: %f" %(ksd1,ksd2,y1,y2,newa1)
    # Computing KS distance for newa1
    ha0N = harmonic(newa1,N)
    modelled[0] = pow(1.0,-newa1)/ha0N
    cummod[0] = pow(1.0,-newa1)/ha0N
    maxdist = abs(cumobs[0]-cummod[0])
    for i in arange(1.0,N-1,1.0):
      modelled[i] = pow(i+1.0,-newa1)/ha0N     
      cummod[i] = cummod[i-1]+(pow(i+1.0,-newa1)/ha0N)
      if maxdist < abs(cumobs[i]-cummod[i]):
        maxdist = abs(cumobs[i]-cummod[i])
    print progress
    bestKSsofar = maxdist
    # Preparing for a new step to focus (check that 10.0 is not too greedy)
    a1 = newa1 - (progress / 10.0)
    a2 = newa1 + (progress / 10.0)
    # Computing KS distance for a1
    ha0N = harmonic(a1,N)
    modelled[0] = pow(1.0,-a1)/ha0N
    cummod[0] = pow(1.0,-a1)/ha0N
    maxdist = abs(cumobs[0]-cummod[0])
    for i in arange(1.0,N-1,1.0):
      modelled[i] = pow(i+1.0,-a1)/ha0N     
      cummod[i] = cummod[i-1]+(pow(i+1.0,-a1)/ha0N)
      if maxdist < abs(cumobs[i]-cummod[i]):
        maxdist = abs(cumobs[i]-cummod[i])
    ks1 = maxdist
    # Computing KS distance for a2
    ha0N = harmonic(a2,N)
    modelled[0] = pow(1.0,-a2)/ha0N
    cummod[0] = pow(1.0,-a2)/ha0N
    maxdist = abs(cumobs[0]-cummod[0])
    for i in arange(1.0,N-1,1.0):
      modelled[i] = pow(i+1.0,-a2)/ha0N     
      cummod[i] = cummod[i-1]+(pow(i+1.0,-a2)/ha0N)
      if maxdist < abs(cumobs[i]-cummod[i]):
        maxdist = abs(cumobs[i]-cummod[i])
    ks2 = maxdist
    # Putting a1 and a2 in the right order
    if ks1 > ks2:
      oldks1 = ks1
      olda1 = a1
      ks1 = ks2
      a1 = a2
      ks2 = oldks1
      a2 = olda1
    # Printing where we stand presently
    print "step: %d, newa1: %f, bestKSsofar: %f, a1: %f, ks1: %f, a2: %f, ks2: %f" %(step,newa1,bestKSsofar,a1,ks1,a2,ks2)
  print("finished")
  # Plot
  # Recomputing the modelled distribution for newa1
  ind = range(1,N+1,1)
  ha0N = harmonic(newa1,N)
  modelled[0] = pow(1.0,-newa1)/ha0N
  cummod[0] = pow(1.0,-newa1)/ha0N
  for i in arange(1.0,N-1,1.0):
    modelled[i] = pow(i+1.0,-newa1)/ha0N
  figure(1)
  plot(ind,observed,'b-',ind,modelled,'g-',antialiased=True)
  xlabel('rank')
  ylabel('Best fitting Zipf\'s law (green) and observed (blue)')
  grid(True)
  show()

if __name__ == '__main__':
  main(sys.argv)
